/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*	"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.	*
*									*
*     Copyright (C) 1993 Lugaru Software Ltd.  All rights reserved.	*
*									*
* Limited permission is hereby granted to reproduce and modify this	*
* copyrighted material provided that the resulting code is used only in	*
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.				*
************************************************************************/

/* For updating from 6.5 or later to 6.55 only. */

#include "eel.h"

	/* possible return values of in_c_comment */
#define IN_COMMENT	0x100	// general mask for any comment
#define IN_OLD_COMMENT	0x101	// in a /* */-type comment
#define IN_NEW_COMMENT	0x102	// in a //-type comment
#define IN_STRING	0x200	// general mask for any string-like thing
#define IN_STR_CONST	0x201	// inside double quotes
#define IN_CHAR_CONST	0x202	// inside single quotes

just_open(p)
{
	save_var point = p;
	return parse_string(1, "[ \t]*{[ \t]*($|//|/<*>)");
}

is_case(p)	/* is there a case-like statement at position p? */
{
	if (!index("cdp", character(p)))
		return 0;
	save_var point = p;
	return parse_string(1, "(case|default|public|private|protected)"
			    "[^a-zA-Z0-9_]");
}

is_switch(p)	/* is there a switch statement at position p? */
{
	if (!index("sc", character(p)))
		return 0;
	save_var point = p;
	return parse_string(1, "(switch|class)[^a-zA-Z0-9_]");
}

in_c_comment(lookback)		// is point inside a c comment, string, etc?
{
	int i = scan_for_string();

	if (i || !lookback)	// check this line first
		return i;
	save_var narrow_start = point - lookback;
	save_var point;
	while (re_search(RE_REVERSE | RE_FASTEST, "/<*>|<*>/")) {
		point++;
		i = scan_for_string();
		if (!i || i == IN_OLD_COMMENT)	// did we find "/*"?
			return curchar() == '*' ? IN_OLD_COMMENT : 0;
	}
	return 0;		// not found, assume we're not in comment
}

when_restoring()
{
	int tflag = 0, i, j;
	char *s, teachname[FNAMELEN];

	after_loading(); /* call early, so we can display errors, etc. */
	putenv("EPSRUNS=Y");	/* for catching nested Epsilons (in DOS) */
	root_keys = reg_tab;
		/* now set buffer-specific variables */
	word_pattern.default = default_word;
	_char_class.default = _def_char_class;
	_case_map.default = _def_case_map;
	_srch_case_map.default = _def_srch_case_map;
	_display_class.default = _std_disp_class;
	_recursion_level = 0;
	i = monochrome ? _our_mono_scheme : _our_color_scheme;
	if (!i)
		i = find_index(monochrome ? "standard-mono"
					  : "standard-color");
	color_scheme = i;
	use_default = 1;		/* default is fundamental mode */
	fundamental_mode();
	use_default = 0;
	apply_defaults();	/* apply defaults to init buffer & window */
	_len_def_mac = 0;	/* at start, not defining a macro */
	run_by_mouse = 0;	/* assume cmds come from kbd */

	strcpy(_docname, "edoc");
	strcpy(teachname, "eteach");
	early_init();

	for (i = 1; i < argc && argv[i][0] == '-'; i++)
		switch (argv[i][1]) {
		case 'f':
			switch (argv[i][2]) {
			case 's':
				set_swapname(argv[i] + 3);
				break;
			case 'h':
				set_shrinkname(argv[i] + 3);
				break;
			case 'd':
				strcpy(_docname, argv[i] + 3);
				break;
			case 't':
				strcpy(teachname, argv[i] + 3);
				break;
			default:
				usage(argv[i]);
			}
			break;
		case 't':
			tflag = 1;
			break;
		case 'r':
		case 'l': break;
		case 'h': case '?': usage("");
		case 'p':
			if (!argv[i][2])
				preserve_session = !preserve_session;
			else if (isdigit(argv[i][2]))
				preserve_session = numtoi(argv[i] + 2);
			else
				cmd_line_session_file = argv[i] + 2;
			break;
		case 'd':
			s = index(argv[i] + 2, '=');
			if (s) {
				*s = 0;
				if (j = find_index(argv[i] + 2)) {
					set_var_val(j, s + 1, 1);
					break;
				}
			}			/* fall through */
		default:
			usage(argv[i]);
		}			/* i is 1st file argument */
	term_init();
#include "keyext.h"
	try_calling("key-setup");
	middle_init();
	init_menu();
	if (tflag) {
		term_cmd_line();
		to_buffer("myteach");
		if (s = lookpath(teachname))
			do_file_read(s, strip_returns.default);
		else
			error("Can't find tutorial. Install first.");
		filename = "myteach";
		preserve_session = 0;
	} else {
		if (i >= argc && load_from_state) {
			load_session();
			term_cmd_line();
		} else {
			term_cmd_line();
			load_cmd_line_files(i);
		}
		delete_dull_buffer("startup");
	}
	try_calling("start-up");
	for (i = 1; i < argc && argv[i][0] == '-'; i++)
		switch (argv[i][1]) {
			case 'l': load_commands(argv[i] + 2); break;
			case 'r': try_calling(argv[i] + 2); break;
		}
}

command exit() on cx_tab[CTRL('C')]
{
	char resp[80];

	iter = 0;
	if (!has_arg)
		error_if_input(0);
#ifdef MSDOS
	if (another_process()) {
		get_string(resp, "A subprocess is running.  Kill it? [y]");
		if (tolower(*resp) == 'n') {
			tiled_only();
			locate_window(PROCBUF, "");
			point = size();
			say("You must kill this subprocess before exiting.");
			return;
		}
		exit_process();
	}
#endif
	if (!has_arg && unsaved_buffers()) {
		do_bufed(0);
		remove_line_highlight();
		maybe_refresh();
		get_string(resp, "Buffers not saved.  Exit? [n]");
		if (tolower(*resp) != 'y') {
			make_line_highlight();
			bufed_msg();
			return;
		}
		remove_bufed();
		maybe_refresh();
	}
	if (!has_arg)
		save_session();
	sayput("");
	finish_up();
	term_position(0, screen_lines - 1);
	term_mode(0);
	if (restore_blinking_on_exit)
		vid_bright_back(0);
	when_exiting();
	leave();
}

check_dates(save) /* compare this file's time info w/ version on disk */
{		/* 0 means ok, 1 means cancel save */
	struct file_info now, *then = &file_info;
	int ret = 1, temp, orig = bufnum;

	if (!want_warn || !filename || then->check_type != CHECK_FILE
			|| check_file(filename, &now) != CHECK_FILE)
		return 0;
	update_readonly_warning(&now);
	if (!compare_dates(&now, then))
		return 0;
	tiled_only();
	switch_to_buffer(bufnum);
	maybe_ding(bell_on_date_warning);
	temp = tmp_buf();
	save_var _view_right = 0, _view_bottom = screen_lines - 22;
	while (!user_abort) {
		show_date_warning(temp, filename, then, &now, want_warn,
						modified, save);
		bufnum = orig;
		view_buf(temp, 0);
		if (user_abort)
			break;
		getkey();
		key = toupper(key);
		if (key == 'C') {
			refresh();
			split_window_vertically();
			window_number--;	/* to 1st window */
			find_in_other_buf(filename, strip_returns);
			filename = "";
			error("This buffer has disk version.");
			break;		/* abort out of calling func */
		} else if (key == 'R') {
			refresh();
			if (!read_file(filename, strip_returns))
				say("%s: Read from disk.", filename);
			break;
		} else if (key == (save ? 'S' : 'I')) {
			refresh();
			ret = 0;
			break;
		} else if (key == '!')
			want_warn = !want_warn;
		say("");
	}
	buf_delete(temp);
	return user_abort ? 1 : ret;
}

/* read the specified file into the current buffer */
read_file(tmp, strip)
char *tmp;
{
	int err;

	err = do_file_read(tmp, strip);
	if (err == 2) {
		zap(bufname);
		modified = 0;
		say("%s: New file.",tmp);
	} else if (err) {
		file_error(err, tmp, "read error");
		maybe_ding(bell_on_read_error);
	} else if (file_info.check_type == CHECK_FILE
				&& (file_info.attr & ATTR_READONLY)) {
		if (readonly_warning & 1)
			say("%s: File is read-only.", tmp);
		if (readonly_warning & 2) {
			_buf_readonly = 2;	// set from file
			call_on_modify = 1;
		}
		maybe_ding(readonly_warning & 4);
	} else
		update_readonly_warning(&file_info);
	make_mode();
	return err;
}

update_readonly_warning(p)  // if file's no longer readonly, and we made
struct file_info *p;	// buffer readonly because it was, make writeable
{
	if (p->check_type == CHECK_FILE && !(p->attr & ATTR_READONLY)
	    && (_buf_readonly == 2)) {
		_buf_readonly = 0;	// if set from file, reset
		make_mode();
	}
}

load_menu()
{
	char buf[FNAMELEN], *s;
	spot sp;

	if (!exist(MENU_BNAME)) {
		if (!(s = lookpath(MENU_FILE)))
			error("Can't find %s", MENU_FILE);
		strcpy(buf, s);
		absolute(buf);
		save_var bufnum = zap(MENU_BNAME);
		if (err_file_read(buf, NORMSTRIP))
			quick_abort();
		tab_size = 100;		// hide second column
		menu_window = 0;
		modify_region(1, MRCOLOR, color_class menu_bar_highlight);
		menu_line_spot = sp = alloc_spot();
		add_region(sp, sp, color_class submenu_highlight, REGLINE);
		submenu_highlight(0);
		save_var abort_searching = 0;
		for (point = 0; re_search(1, "^#.*\n"); )
			delete(matchstart, point);
	}
}

command rebuild_menu()
{
	save_var bufnum;
	if (permanent_menu)
		drop_perm_menu();
	if (exist(MENU_BNAME))
		delete_buffer(MENU_BNAME);
	if (permanent_menu)
		make_perm_menu();
}

one_menu_binding(func, needspace)	/* add binding to a menu item */
char *func;
{
	int i;
	char b[80];

	if (*func == '#' || needspace < 0)
		return 0;
	*b = 0;
	i = find_index(func);
	if (!i || !find_bindings(b, i, mode_keys, root_keys, "", 0))
		return 0;
	for (i = 0; i < needspace; i++)
		insert(' ');
	stuff(b);
	return strlen(b);
}

is_menu_here()	// has mouse moved onto (possibly not displayed) menu bar?
{
	return still_on_menu_bar() && (auto_menu_bar && mouse_pixel_y <= 0
				       && y_pixels_per_char() > 1
				       || permanent_menu);
}

do_show_menu(num)
{
	char cmd[300];
	int sub, doit;

	save_var window_handle;
	save_var abort_searching = 0;
	save_var mouse_mask |= MASK_MOVE | MASK_UP;
	if (!IS_MOUSE_KEY(key))
		save_var menu_stays_after_click = 1;
	sub = show_menu_loop(num, cmd);
	restore_vars();
	doit = !exit_menu_cancel();
	finish_menu_selection(cmd, sub, doit);
	if (!echo_line)	// if echo line at top, clear menu from it
		note("");
	if (doit)
		run_menu_selection(cmd, shift_pressed());
}

show_menu_loop(num, cmd)
char *cmd;
{
	int sub = 0, line = 0, max, isdir, oldsub, orignum;

	max = max_submenus();
	if (IS_MOUSE_KEY(key))
		ungot_key = key;
	do {
		oldsub = sub;
		sub = to_sub_line(sub, num, line);
		if (getkey() == MOUSE_MOVE) {
			refresh();
			while (getkey() == MOUSE_MOVE)
				;
		}
		isdir = 1;		// assume it's a direction key
		switch (get_direction()) {
			case BTOP:	line--; break;
			case BBOTTOM:	line++; break;
			case BLEFT:	num--, line = 0; break;
			case BRIGHT:	num++, line = 0; break;
			default:	isdir = 0; break;
		}
		if (isalpha(key))
			line = do_menu_search(key);
		else if (IS_MOUSE_DOWN(key)) {
			if (oldsub && mouse_y <= 0)	// maybe drop menu
				orignum = count_submenus(mouse_x) - 1;
			else
				orignum = -2;	// upclick on bar won't drop
			sub = while_mouse_pressed(sub);
			num = after_mouse_pressed(sub, num, orignum);
			line = 0;
		} else if (reg_tab[key] == (short) help) {
			get_menu_cmd(cmd, sub);
			if (*cmd && find_index(cmd))
				help_on_command(cmd);
			isdir = 1;
		}
		num = (num + max) % max;
		line = (line + window_height) % window_height;
	} while (isdir || isalpha(key) || IS_MOUSE_KEY(key));
	if (!exit_menu_doit() && !exit_menu_cancel()) {
		ungot_key = key;
		key = ESC;
	}
	return sub;
}

mouse_move_scroll_box()		// grabbed box in center of scroll bar
{
	int i, pix, max, offset, last = -1;
	char *p = display_characters + SCROLLCHARSTART + 2, old = *p;

	save_var mouse_display = 0, mouse_auto_on = 0;
	*p = p[1];
	pix = y_pixels_per_char();
	max = (MAX(text_height(), 5) - 4) * pix;
	i = mouse_pixel_y - (w_top + 2) * pix;
	offset = point - size() * i / max;
	while (mouse_pressed() && !user_abort) {
		if (char_avail())
			getkey();
		i = mouse_pixel_y - (w_top + 2) * pix;
		point = size() * i / max + offset;
		if (point > 0 && goal_column >= 0)
			move_to_column(goal_column);
		window_start = prev_screen_line(text_height() / 2);
		if (point != last)
			refresh();
		last = point;
	}
	*p = old;
}

#ifdef UNIX		/* i-search redoes cmd unless \r or ENDER */
#define ENDER	'\r'		/* only \r: return replaces ESC */
#else
#define ENDER	ESC		/* allow both */
#endif

#define FHERE	1	/* s_code: where to start searching from */
#define FORIG	2
#define FMATCH	3	/* from last match */
#define FMADD	4	/* same, but indicates we added a char this time */
#define FMREGEX	5	/* same, but in regex mode */

search_read(trystr, prmpt, flags)
char *trystr, *prmpt;
{
	int lastok, goodchars, oldflags, mstart, origwin = window_handle;
	int editbuf, pbuf, pwin, editwin, s_code, okpoint, orig, ostart;

	if (_doing_input == DI_SEARCH)
		error_if_input(0);
	save_var iter, has_arg, this_cmd, prev_cmd;
	save_var _doing_input = DI_SEARCH;
	save_var last_index;
	goodchars = lastok = 0;	/* 2 versions of len of last good string */
				/* (only lastok is set by getting default) */
	orig = okpoint = mstart = point;
	ostart = window_start;

	editbuf = tmp_buf();		/* make buffer for cmd line editing */
	pwin = add_popup(0, echo_line, screen_cols - 1, 1, 0,
			pbuf = tmp_buf());	/* prompt in window */
	stuff(trystr);
	display_column = 0;
	system_window = 1;	/* don't let user switch to it */
	text_color = color_class echo_area;

	editwin = add_popup(1, echo_line, screen_cols - 1, 1,
			0, editbuf);	/* window on top of prompt window */
	text_color = color_class echo_area;
	_display_class = _echo_disp_class;
	display_column = 0;

	for (;;) {
		do {
			show_search_prompt(prmpt, flags, origwin,
					   pwin, editwin);
			getkey();
		} while (key == MOUSE_MOVE);
		if (IS_MOUSE_KEY(key) && mouse_y == echo_line)
			flags &= ~INCR;
		if (key == CTRL('^'))
			full_getkey("", CTRLIFY_KEY);
		window_handle = editwin;
		s_code = 0;
		oldflags = flags;
		flags &= ~(BAD | FAILING | SONCE);	/* assume ok */
		if (is_search_dir()) {
			window_handle = origwin;
			point = okpoint;
			window_handle = editwin;
			if ((key == CTRL('R')) != ((flags & REVERSE) != 0)) {
				flags ^= REVERSE;	/* switch dir */
				if (goodchars)
					flags |= SONCE;
			} else if (!size()) {
				if (flags & UNKNOWN)
					flags = known_flags(flags);
get_defaults:			stuff(default_search_string(flags & REGEX));
				lastok = size();
			} else if (oldflags & FAILING)
				flags |= FAILING; /* maybe search again */
			else
				flags |= SONCE;
			if ((flags & (INCR | SONCE)) && !(flags & FAILING))
				flags |= SONCE, s_code = FHERE;
		} else if (set_flag_modes(&flags)) {
			if (flags & INCR) /* if modes changed, search again */
				s_code = FMATCH;
		} else if (key == CTRL('O')) {
			flags ^= INCR;
			point = size();
			if (flags & INCR)
				s_code = FMATCH;
		} else if (reg_tab[key] == (short) help)
			help_on_command(help_on_search(flags));
		else if (IS_ENTER_KEY(key)) {
			if (oldflags & SONCE)
				flags |= SONCE;
			break;
		} else if (reg_tab[key] == (short) backward_delete_character
						&& (flags & INCR)) {
			run_topkey();
			window_handle = editwin;
			if (!(oldflags & FAILING) || size() <= lastok)
				s_code = FORIG;
			else		/* don't bother searching */
				flags |= FAILING;
		} else if (key == abort_key && (oldflags &
					((flags & INCR) ? FAILING : SONCE))) {
			if (flags & INCR)
				delete(goodchars, size());
		} else if (user_abort || (flags & INCR) && exit_incr())
			break;
		else {
			flags &= ~UNKNOWN;
			if (key == '\t')
				normal_character();
			else
				run_topkey();
			window_handle = editwin;
			if (flags & INCR)
				if (flags & (REGEX | WORD))
					s_code = FMREGEX;
				else if (oldflags & FAILING)
					flags |= FAILING;
				else
					s_code = FMADD;
		}
		user_abort = 0;
		grab(0, FNAMELEN - 1, trystr);
		if (!s_code)
			continue;
		window_handle = origwin;
		if (s_code == FORIG || s_code == FMREGEX)
			point = orig;
		else if (s_code >= FMATCH) {
			point = mstart;
			if (flags & REVERSE)
				point++;
		}
		switch (do_searching(flags, trystr)) {
			case DSBAD:
				flags = BAD | (flags & ~INCR);
			case 0:	flags |= FAILING;
				point = okpoint;
				if (in_macro()) {
					show_search_prompt(prmpt, flags,
						origwin, pwin, editwin);
					refresh();
					goto end;
				} else if (!(oldflags & FAILING))
					maybe_ding(bell_on_search);
				break;
			case 1:	okpoint = point;
				mstart = matchstart;
				window_handle = editwin;
				goodchars = lastok = size();
				break;
			case DSABORT:
				point = okpoint;
				window_handle = editwin;
				if ((flags & INCR) && s_code == FMADD)
					delete(size() - 1, size());
				break;
		}
	}
	window_handle = origwin;
	if (key == abort_key || user_abort)
		point = orig, window_start = ostart;
	else {
		window_handle = editwin;
		if (size()) {
			_last_was_regex = ((flags & REGEX) != 0);
			*(_last_was_regex ? &_default_regex_flags
				: &_default_flags) = (flags & (SMODES | INCR));
			grab(0, FNAMELEN - 1,
			     default_search_string(_last_was_regex));
		}
		if ((flags & INCR) && !IS_ENTER_KEY(key) && key != ENDER)
			ungot_key = key;
	}
end:	window_handle = origwin;
	remove_window(editwin);
	remove_window(pwin);
	buf_delete(editbuf);
	buf_delete(pbuf);
	if ((flags & FAILING) && in_macro())
		quick_abort();
	check_abort();
	say("");
	maybe_refresh();	/* put cursor back in prev window */
	return flags;
}

string_replace(str, with, flags)
char *str, *with;
{
	spot orig = alloc_spot(), prev = alloc_spot(), oldmark;
	int skip = 0, show = 0, ret, evershow = 0, found = 0, changed = 0;

	for (;;) {
		skip = 0;
		*prev = point;
		ret = do_searching(flags, str);
		if (!ret || ret == DSBAD) {
			point = *orig;
			break;
		} else if (ret == DSABORT)
			break;
		found++;
ask:		if (show) {
			show_replace(str, with, (flags & ~INCR));
			evershow = 1;
			show = 0;
		}
		if (!(flags & QUERY)) goto do_replace;
		maybe_refresh();		/* show the selection */
		getkey();
		if (key == CTRL('^'))
			full_getkey("", CTRLIFY_KEY);
		user_abort = 0;
		if (key == '?' || key == 'h'
					|| root_keys[key] == (short) help) {
			help_on_command("query-replace");
			show = 1;
			point = *prev;
			continue;
		} else if (user_abort || key == abort_key) {
			point = *orig;
			ret = DSABORT;
			goto done;
		}
		switch(tolower(key)) {
			case '^': point = *prev;
			case '\b':
			case 'n': skip = 1;	/* go to the next match */
			case ' ':
			case 'y':		/* do the replace */
			case '!':
			case ',':
			case '.': break;
			case CTRL('R'):
				say("");
				oldmark = alloc_spot();
				*oldmark = mark;	/* save mark */
				mark = point - strlen(skip? with: str);
					/* set temporary mark before text */
				recursive_edit();
				mark = *oldmark;	/* restore mark */
				free_spot(oldmark);
				show = 1;
				continue;
			default:	if (set_flag_modes(&flags)) {
						show = 1;
						goto ask;
					}
					if (!IS_ENTER_KEY(key))
						ungot_key = key;
			case ESC:	ret = 1; goto done;
		}
do_replace:	if (!skip) {
			if (flags & REGEX)
				fromorig(with);
			else {
				delete(point - strlen(str), point);
				stuff(with);
			}
			changed++;
		}
		if (flags & QUERY)
			switch (key) {
			case ',': case '^':	skip = 1; goto ask;
			case '.':		ret = 1; goto done;
			case '!':		flags &= ~QUERY; show = 1;
			}
	}
done:
	if (ret == DSABORT)
		say("Replacement aborted");
	else if (ret == DSBAD) {
		show_replace(str, with, (flags & ~INCR) | BAD);
	} else if ((evershow || (flags & INCR)) && !in_macro())
		say("Replaced %d of %d.", changed, found);
	free_spot(orig);
	free_spot(prev);
	return ret;
}

set_readonly_from_file()
{
	if (file_info.check_type == CHECK_FILE &&
			(file_info.attr & ATTR_READONLY) &&
			(readonly_warning & 2)) {
		_buf_readonly = 2;	// set from file
		call_on_modify = 1;
	}
}

rest_sess_wins(b, wins, skip, def_buf)
{
	int nums[10], i, err = 0, bnum, must_default;
	char buf[FNAMELEN];
	struct window_info w;

	for (i = 0; i < wins; i++) {
		if (grab_numbers(b, nums) < 4)
			err = 6;
		w.left = nums[0];
		w.top = nums[1];
		w.right = nums[2];
		w.bottom = nums[3];
		if (grab_numbers(b, nums) < 4)
			err = 7;
		w.textcolor = nums[0];
		w.hbordcolor = nums[1];
		w.vbordcolor = nums[2];
		w.titlecolor = nums[3];
		if (grab_numbers(b, nums) < 2)
			err = 8;
		w.borders = nums[0];
		w.other = nums[1];
		grab_line(b, buf);
		w.point = -1;
		w.dpoint = -1;
		bnum = name_to_bufnum(buf);
		must_default = !bnum;
		grab_line(b, buf);
		w.prevbuf = name_to_bufnum(buf);
		w.bufnum = bnum ? bnum : w.prevbuf ? w.prevbuf :
			(bnum = give_prev_buf()) ? bnum : def_buf;
		if (grab_numbers(b, nums) < 3)
			err = 9;
		w.dcolumn = nums[1];
		if (!must_default) {
			w.point = nums[0];
			w.dpoint = nums[2];
		}
		if (err)
			break;
		if (!skip) {
			window_create(!i, &w);
			fix_window_start();
		}
	}
	nums[1] = screen_lines - 1;	/* echo line */
	nums[2] = 0;			/* want perm menu? */
	nums[3] = 0;			/* top avoid lines */
	nums[4] = 1;			/* bottom avoid lines */
	nums[5] = 0;			/* mode line at top */
	nums[6] = 0;			/* want perm scroll bar? */
	if (!err && !grab_numbers(b, nums))
		err = 10;
	if (err)
		one_window();
	else if (!skip) {
		select_low_window(nums[0], nums[3], nums[4]);
		echo_line = nums[1];
		restore_menu(nums[2]);
		mode_line_at_top = nums[5];
		tiled_scroll_bar = nums[6];
		if (tiled_scroll_bar)
			border_right = 1;
	}
	return err;
}

tag_c_func()	// tag this, if it's a function definition, or maybe decl
{		// return 0: go back & try again, 1: continue, 2: also tag
	int ret = 2;

	if (!move_level(1, "(", ")"))
		return 0;	/* after identifier, skip over arg list */
	c_fwd_space();
	if (parse_string(1, "[A-Za-z{]"))	// func definition
		ret = skip_param_decls();
	else if (!parse_string(1, "[,;]"))	// not even a declaration
		ret = 0;
	else if (!(tag_declarations & 1))	// a declaration, don't tag it?
		ret = 1;
	if (ret)
		tag_extern_decl = 0;
	return ret;
}

skip_param_decls()	// move past param decls in func def to just before "{"
{		// return 0: go back & try again, 1: continue, 2: also tag
	int orig = point;	// good if no decls, or they end with ";"

	if (curchar() == '{' || search(1, "{") && semi_before(--point))
		return 2;	// had ; just before {, good decls
	point = orig;		// these weren't param decls, go back
	return 1;
}

semi_before(p)	// is there a semicolon just before p (ignoring comments)?
{
	save_var point = p;
	return !cind_try_space() && character(point - 1) == ';';
}

tag_c_variable()	// is this a var definition?
{
	if (parse_string(1, "([) \t\n]|<[>[^]\n]*<]>)*[,;=]")) {	
		point = matchend;	// ordinary variable definition
		if (character(point - 1) == '=')
			skip_initializer();
	} else if (parse_string(1, "[ \t\n]*<)>+[ \t\n]*!<(>"))
		re_search(1, "![;{]");	// func ptr definition
	else
		return 0;
	return 1;
}

#ifdef OS2
#include "lowlevel.h"

int _inside_size;	/* screen size while in epsilon */
int _outside_size;	/* size during push or at exit */
short _last_video_mode = -1;		/* last video mode we set */

short _video_adapter;
#define IS_EGA		(_video_adapter == 2)
#define IS_VGA		(_video_adapter == 3 || _video_adapter >= 7)

_find_outside()	/* set variables for screen size outside of epsilon */
{
	short data[2];

	_last_video_mode = screen_mode;
	_outside_size = vid_get_mode();
	data[0] = sizeof(data);
	data[1] = -1;
	os2call("VIOCALLS", 0, "VIOGETCONFIG", 4,
		0, get_pointer(data, SEGMENT),
		get_pointer(data, OFFSET), 0, 0, 0);
	_video_adapter = data[1];
	if (IS_EGA || IS_VGA)
		selectable_colors = 256;
}

video_mode_80x43(active)
{
	if (active != 1)
		return IS_EGA || IS_VGA;
	return vid_set_mode(80, 43, 0, 0) ? 1
		: vid_set_mode(80, 43, 720, 350) ? 1
		: vid_set_mode(80, 43, 640, 350);
}

video_vga_mode(active, lines, vres)
{
	if (active != 1)
		return IS_VGA;
	return vid_set_mode(80, lines, 0, 0) ? 1
		: vid_set_mode(80, lines, 720, vres) ? 1
		: vid_set_mode(80, lines, 640, vres);
}

#define MAKE_SIZE(cols, lines)	((lines) * 1000 + (cols))
struct ModeData _video_mode_data;

vid_set_mode(cols, lines, hres, vres)
{
	if (vid_get_mode() == MAKE_SIZE(cols, lines))
		return 1;
	term_clear();
	_video_mode_data.col = cols, _video_mode_data.row = lines;
	if (hres) _video_mode_data.hres = hres;
	if (vres) _video_mode_data.vres = vres;
	vid_xxx_call("VIOSETMODE", &_video_mode_data);
	term_clear();
	term_position(0, 0);
	leave_blank = 1;	/* note that screen's clear */
	vid_bright_back(1);
	return vid_get_mode() == MAKE_SIZE(cols, lines);
}

vid_bright_back(bright)	// set high background colors to bright or blinking
{
	short params[3];

	params[0] = sizeof(params);
	params[1] = 2;
	params[2] = bright;
	vid_xxx_call("VIOSETSTATE", params);
}
#endif

window_to_fit(max)	/* adjust window size to display whole buffer */
{
	window_height = max;		/* set to max size */
	build_window();			/* construct it internally */
/* if whole buffer fits in window, shrink by # of blank lines at bottom */
	if (!window_start && window_extra_lines() > 0)
		window_height -= window_extra_lines();
}

when_loading()
{
	if (find_index("load-changes"))
		load_from_path("store");
}
